import json
import os

from websockets.sync.client import connect, ClientConnection


class VTSParameterData:
    def __init__(self, id: str, value: any):
        self.id = id
        self.value = value


class VTSPluginInfo:
    def __init__(self, plugin_name: str, developer: str, authentication_token_path: str, plugin_icon: str = None):
        self.plugin_name = plugin_name
        self.developer = developer
        self.authentication_token_path = authentication_token_path
        self.plugin_icon = plugin_icon


class VTSWebSocket:
    def __init__(self, ws_uri: str, plugin_info: VTSPluginInfo):
        self.ws_uri = ws_uri
        self.plugin_info = plugin_info
        self.req_msg = {
            "apiName": "VTubeStudioPublicAPI",
            "apiVersion": "1.0",
        }

        self.client: ClientConnection | None = None
        self.vts_token = None

    def close(self):
        assert self.client is not None
        self.client.close()

    def request(self, api: str, req_id: str = None, data: dict = None, response: bool = True) -> any:
        assert self.client is not None
        req = {**self.req_msg, "requestID": api if req_id is None else req_id, "messageType": api}
        if data is not None:
            req = {**req, "data": data}
        self.client.send(json.dumps(req, ensure_ascii=False, skipkeys=True))
        res = self.client.recv()
        if response:
            return json.loads(res)

    def connect(self):
        self.client = connect(uri=self.ws_uri)
        res = self.request('APIStateRequest')
        if res['data']['currentSessionAuthenticated'] is True:
            return
        if os.path.exists(self.plugin_info.authentication_token_path):
            with open(self.plugin_info.authentication_token_path, 'r', encoding='utf-8') as f:
                self.vts_token = f.read()
        else:
            self.authenticate_token()

        if self.authenticate():
            return

        self.authenticate_token()
        if self.authenticate() is False:
            raise Exception('Authentication failed')

    def authenticate(self) -> bool:
        assert self.vts_token is not None
        res = self.request('AuthenticationRequest', data={"pluginName": self.plugin_info.plugin_name,
                                                          "pluginDeveloper": self.plugin_info.developer,
                                                          "authenticationToken": self.vts_token})
        return res['data']['authenticated']

    def authenticate_token(self):
        res = self.request('AuthenticationTokenRequest', data={"pluginName": self.plugin_info.plugin_name,
                                                               "pluginDeveloper": self.plugin_info.developer,
                                                               "pluginIcon": self.plugin_info.plugin_icon})
        self.vts_token = res['data']['authenticationToken']
        with open(self.plugin_info.authentication_token_path, 'w', encoding='utf-8') as f:
            f.write(self.vts_token)

    def set_params(self, params: list[VTSParameterData]):
        data = {
            "faceFound": False,
            "mode": "add",
            "parameterValues": [d.__dict__ for d in params],
        }
        self.request('InjectParameterDataRequest', data=data, response=False)

    def set_single_param(self, param: VTSParameterData):
        self.set_params([param])

# ws = VTSWebSocket()
# ws.connect()
# ss = VTSParameterData('MouthOpen', 0.9)
# ws.set_single_param(ss)
# print(json.dumps(ss.__dict__))



import numpy as np
import soundfile as sf

from .analyser import Analyser, DBAnalyser, VowelAnalyser
from .vts_websockets import VTSWebSocket, VTSPluginInfo, VTSParameterData


class VTSAdapter:
    def __init__(self,
                 analyser: Analyser,
                 db_vts_mouth_param: str = 'MouthOpen',
                 vowel_vts_mouth_param: dict[str, str] = None,
                 ws_uri: str = 'ws://localhost:8001',
                 plugin_info: VTSPluginInfo = VTSPluginInfo(plugin_name='pymouth',
                                                            developer='organics',
                                                            authentication_token_path='./pymouth_vts_token.txt',
                                                            plugin_icon=None)
                 ):
        """
        VTubeStudio Adapter.
        :param analyser: 分析仪,必须是 Analyser 的子类
        :param db_vts_mouth_param: 仅作用于`DBAnalyser`, VTS中控制mouth_input的参数, 如果不是默认值请自行修改.
        :param vowel_vts_mouth_param: 仅作用于VowelAnalyser, VTS中控制mouth的参数,这个参数默认是 ('VoiceSilence', 'VoiceA', 'VoiceI', 'VoiceU', 'VoiceE', 'VoiceO')
            如果你的VTS中不是这些默认值，你可以通过这个参数修改以匹配你VTS中关于AIUEO的输入参数
            例如：{
                    'VoiceSilence': 'CustomSilence',
                    'VoiceA': 'CustomA',
                    'VoiceI': 'CustomI',
                    'VoiceU': 'CustomU',
                    'VoiceE': 'CustomE',
                    'VoiceO': 'CustomO'
                }
            这个dict中 key 不能变，value 则设置你自己VTS中关于AIUEO的输入参数

        :param ws_uri: websocket uri 默认：ws://localhost:8001
        :param plugin_info: 插件信息,可以自定义
        """

        if vowel_vts_mouth_param is None:
            vowel_vts_mouth_param = {
                'VoiceSilence': 'VoiceSilence',
                'VoiceA': 'VoiceA',
                'VoiceI': 'VoiceI',
                'VoiceU': 'VoiceU',
                'VoiceE': 'VoiceE',
                'VoiceO': 'VoiceO'
            }

        self.analyser = analyser

        self.db_vts_mouth_param = db_vts_mouth_param
        self.vowel_vts_mouth_param = vowel_vts_mouth_param

        self.ws_uri = ws_uri
        self.plugin_info = plugin_info
        self.vts = VTSWebSocket(ws_uri, plugin_info)

    def __enter__(self):
        self.vts.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.vts.close()

    def __db_callback(self, y: float, data):
        self.vts.set_single_param(VTSParameterData(self.db_vts_mouth_param, y))

    def __vowel_callback(self, vowel_dict: dict, data):
        params = [VTSParameterData(self.vowel_vts_mouth_param[k], v) for k, v in vowel_dict.items()]
        self.vts.set_params(params)

    def action(self,
               audio: np.ndarray | str | sf.SoundFile,
               samplerate: int | float,
               output_device: int,
               finished_callback=None,
               auto_play: bool = True):

        """
        启动分析器开始分析音频数据, 注意:此方法为非阻塞方法,会立即返回
        :param audio: 音频数据, 可以是文件path, 可以是SoundFile对象, 也可以是ndarray
        :param samplerate: 采样率, 这取决与音频数据的采样率, 如果你无法获取到音频数据的采样率, 可以尝试输出设备的采样率.
        :param output_device: 输出设备Index, 这取决与硬件或虚拟设备. 可用 audio_devices_utils.py 打印当前系统音频设备信息
        :param finished_callback: 音频处理完成后,会回调这个方法
        :param auto_play: 是否自动播放音频,默认为True,会播放音频(自动将audio写入指定`output_device`)
        """

        if isinstance(self.analyser, DBAnalyser):
            self.analyser.action_noblock(audio,
                                         samplerate,
                                         output_device,
                                         self.__db_callback,
                                         finished_callback,
                                         auto_play,
                                         block_size=4096)
        elif isinstance(self.analyser, VowelAnalyser):
            self.analyser.action_noblock(audio,
                                         samplerate,
                                         output_device,
                                         self.__vowel_callback,
                                         finished_callback,
                                         auto_play,
                                         block_size=4096)

    def action_block(self,
                     audio: np.ndarray | str | sf.SoundFile,
                     samplerate: int | float,
                     output_device: int,
                     finished_callback=None,
                     auto_play: bool = True):

        if isinstance(self.analyser, DBAnalyser):
            self.analyser.action_block(audio,
                                       samplerate,
                                       output_device,
                                       self.__db_callback,
                                       finished_callback,
                                       auto_play,
                                       block_size=4096)
        elif isinstance(self.analyser, VowelAnalyser):
            self.analyser.action_block(audio,
                                       samplerate,
                                       output_device,
                                       self.__vowel_callback,
                                       finished_callback,
                                       auto_play,
                                       block_size=4096)



import traceback
from abc import ABCMeta
from concurrent.futures import ThreadPoolExecutor

import librosa
import numpy as np
import sounddevice as sd
import soundfile as sf
from dtw import dtw


class Analyser(metaclass=ABCMeta):
    V_A = [[157.28203, -18.084631],
           [184.55794, -70.21922],
           [180.11708, -69.37925],
           [186.3136, -68.28505],
           [184.01038, -62.837513],
           [181.70918, -56.22454],
           [183.93184, -55.787033],
           [177.58957, -55.56417],
           [210.36815, -42.245583]]
    V_I = [[154.23402, - 66.638794],
           [109.41006, -76.42681],
           [104.646484, -79.092224],
           [100.69435, -73.16504],
           [106.442406, -68.50664],
           [104.717026, -67.13239],
           [98.70581, -67.052155],
           [90.54036, -64.939255],
           [96.68744, -4.7852263]]
    V_U = [[177.38, -21.598547],
           [163.56247, -60.48896],
           [162.29306, -65.40919],
           [147.10324, -77.276535],
           [145.81586, -81.49585],
           [141.00285, -80.76966],
           [137.3823, -83.69654],
           [125.162674, -86.36088],
           [182.829, -57.803486]]
    V_E = [[45.139225, -3.552611],
           [82.19922, -65.22143],
           [154.64558, -90.886116],
           [153.86205, -97.447235],
           [158.62253, -77.25729],
           [155.23709, -56.107735],
           [149.8651, -64.127594],
           [145.0244, -55.381966],
           [175.49411, -33.858017]]
    V_O = [[224.24329, -10.556553],
           [193.64792, -22.315527],
           [186.96034, -20.628428],
           [184.96133, -16.40874],
           [186.04208, -22.023905],
           [185.25256, -11.928689],
           [185.24428, -5.041305],
           [181.70624, -3.2469425],
           [131.43114, 3.3550904]]
    V_Silence = [[52.750603, -4.9177713],
                 [84.026146, 21.443743],
                 [77.42841, 2.3937027],
                 [85.50844, 21.711079],
                 [80.75325, 7.308112],
                 [82.622925, 16.251738],
                 [69.67996, 10.19037],
                 [80.5309, 6.884417],
                 [71.37026, 22.482405]]

    def __init__(self, temperature: float = 10.0):
        self.executor = ThreadPoolExecutor(1)
        self.temperature = temperature

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def action_noblock(self,
                       audio: np.ndarray | str | sf.SoundFile,
                       samplerate: int | float,
                       output_device: int,
                       callback,
                       finished_callback=None,
                       auto_play: bool = True,
                       dtype: np.dtype = np.float32,
                       block_size: int = 4096):

        self.executor.submit(self.action_block,
                             *(audio,
                               samplerate,
                               output_device,
                               callback,
                               finished_callback,
                               auto_play,
                               dtype,
                               block_size))

    def action_block(self,
                     audio: np.ndarray | str | sf.SoundFile,
                     samplerate: int | float,
                     output_device: int,
                     callback,
                     finished_callback=None,
                     auto_play: bool = True,
                     dtype: np.dtype = np.float32,
                     block_size: int = 4096):

        stream = None
        try:
            if isinstance(audio, np.ndarray):
                # 声道验证
                if audio.ndim <= 0 or audio.ndim > 2:
                    raise ValueError('Audio channel verification failed. Only single or dual channels are supported.')

                stream = sd.OutputStream(samplerate=samplerate,
                                         blocksize=block_size,
                                         device=output_device,
                                         channels=audio.ndim,  # if data.shape[1] != channels:
                                         dtype=dtype).__enter__() if auto_play else None

                datas = split_list_by_n(audio, block_size)
                for data in datas:
                    self.play(callback, data, samplerate, stream)

            elif isinstance(audio, str):
                with sf.SoundFile(audio) as f:
                    stream = sd.OutputStream(samplerate=f.samplerate,
                                             blocksize=block_size,
                                             device=output_device,
                                             channels=f.channels,
                                             dtype=dtype).__enter__() if auto_play else None
                    while True:
                        data = f.read(block_size, dtype=dtype)
                        if not len(data):
                            break
                        self.play(callback, data, samplerate, stream)

            elif isinstance(audio, sf.SoundFile):
                stream = sd.OutputStream(samplerate=audio.samplerate,
                                         blocksize=block_size,
                                         device=output_device,
                                         channels=audio.channels,
                                         dtype=dtype).__enter__() if auto_play else None

                while True:
                    data = audio.read(block_size, dtype=dtype)
                    if not len(data):
                        break
                    self.play(callback, data, samplerate, stream)

        except Exception:
            traceback.print_exc()
        finally:
            if stream is not None:
                stream.__exit__()

            if finished_callback is not None:
                finished_callback()

    def play(self, callback, data: np.ndarray, samplerate: int | float, stream: sd.OutputStream):
        if stream is not None:
            stream.write(data)
        callback(self.process(data, samplerate), data)

    def process(self, data: np.ndarray, samplerate: int | float):
        pass

    def _audio2vowel(self, audio_data: np.ndarray, samplerate: int | float) -> dict[str, float]:
        audio_data = channel_conversion(audio_data)

        # TODO 这里可能要做人声滤波 , 人声分离比较复杂且耗费性能，暂时不提供支持
        # 对线性声谱图应用mel滤波器后，取log，得到log梅尔声谱图，然后对log滤波能量（log梅尔声谱）做DCT离散余弦变换（傅里叶变换的一种），然后保留第2到第13个系数，得到的这12个系数就是MFCC
        n_fft = get_n_fft(audio_data.size, samplerate)
        mfccs = librosa.feature.mfcc(y=audio_data, sr=samplerate, n_fft=n_fft, dct_type=1, n_mfcc=3)[1:].T
        # 过短的音频会导致无法比较，直接按无声处理
        if mfccs.shape[0] < 5:
            return {
                'VoiceSilence': 1,
                'VoiceA': 0,
                'VoiceI': 0,
                'VoiceU': 0,
                'VoiceE': 0,
                'VoiceO': 0,
            }

        # 通过DTW(动态时间规整算法) 计算 当前帧窗与元音帧窗的距离，值越小越相似
        si = dtw(self.V_Silence, mfccs, distance_only=True).normalizedDistance
        a = dtw(self.V_A, mfccs, distance_only=True).normalizedDistance
        i = dtw(self.V_I, mfccs, distance_only=True).normalizedDistance
        u = dtw(self.V_U, mfccs, distance_only=True).normalizedDistance
        e = dtw(self.V_E, mfccs, distance_only=True).normalizedDistance
        o = dtw(self.V_O, mfccs, distance_only=True).normalizedDistance

        # log = f"Silence:{si:f}, A:{a:f}, I:{i:f}, U:{u:f}, E:{e:f}, O:{o:f}"
        # print(log)
        # 对距离取负，值越大越相似，再对相似度进行softmax，取相似度的概率分布。这里 temperature 取大值降低置信度，可以使输出元音的整体概率更平滑，口型更加真实。
        r = np.array([-1 * i for i in [si, a, i, u, e, o]])
        r = softmax(r, temperature=self.temperature).tolist()
        # print([f"{t:f}" for t in r])

        res = {
            'VoiceSilence': r[0],
            'VoiceA': r[1],
            'VoiceI': r[2],
            'VoiceU': r[3],
            'VoiceE': r[4],
            'VoiceO': r[5],
        }

        return res


class DBAnalyser(Analyser):
    def __init__(self, temperature: float = 10.0):
        super().__init__(temperature=temperature)

    def process(self, data: np.ndarray, samplerate: int | float):
        return self._audio2db(data, samplerate)

    def _audio2db(self, audio_data: np.ndarray, samplerate: int | float) -> float:
        res = self._audio2vowel(audio_data, samplerate)
        vs = res['VoiceSilence']
        max = np.max([res[k] for k in res.keys()])
        return max if max != vs else 0


class VowelAnalyser(Analyser):
    def __init__(self, temperature: float = 10.0):
        super().__init__(temperature=temperature)

    def process(self, data: np.ndarray, samplerate: int | float):
        return self._audio2vowel(data, samplerate)


def channel_conversion(audio: np.ndarray):
    # 如果音频数据为立体声，则将其转换为单声道
    if audio.ndim == 2 and audio.shape[1] == 2:
        return audio[:, 0]
    return audio


def split_list_by_n(list_collection, n):
    """
    将集合均分，每份n个元素
    :param list_collection:
    :param n:
    :return:返回的结果为评分后的每份可迭代对象
    """
    for i in range(0, len(list_collection), n):
        yield list_collection[i: i + n]


n_fft_ref_list = [128, 512, 1024, 2048, 4096]


def get_n_fft(audio_block_size: int, sample_rate: int, milliseconds: int = 23) -> int:
    """
    计算FFT，计算给定时间和采样率内最适合的FFT。
    一般认为每23毫秒左右为发音的一帧，对每一帧做 FFT（快速傅里叶变换），傅里叶变换的作用是把时域信号转为频域信号
    :param audio_block_size:
    :param sample_rate:
    :param milliseconds:
    :return:
    """
    n_fft = sample_rate * (milliseconds / 1000)
    if n_fft > audio_block_size:
        return audio_block_size
    return min(n_fft_ref_list, key=lambda n_fft_ref: abs(n_fft_ref - n_fft))


def softmax(x: np.ndarray, temperature: float = 1.0) -> np.ndarray:
    """
    带温度参数的softmax函数
    temperature > 1：使分布更平滑（降低置信度）
    temperature < 1：使分布更尖锐（提高置信度）
    temperature = 1：标准softmax

    参数:
    x -- 输入向量或矩阵(numpy数组)
    temperature -- 温度参数(正数，默认为1.0)

    返回:
    s -- softmax计算结果(与x形状相同)

    异常:
    ValueError -- 如果温度<=0
    """
    if temperature <= 0:
        raise ValueError("Temperature must be positive")

    # 防止数值溢出，减去最大值
    x = x / temperature
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0, keepdims=True)



import pyaudio


def get_audio_devices():
    p = pyaudio.PyAudio()
    devices = []
    for i in range(p.get_device_count()):
        # print(p.get_device_info_by_index(i).get('name'))
        devices.append(p.get_device_info_by_index(i))
    return devices


def get_audio_input_devices():
    devices = []
    for item in get_audio_devices():
        if item.get('maxInputChannels') > 0:
            devices.append(item)
    return devices


def get_audio_output_devices():
    devices = []
    for item in get_audio_devices():
        if item.get('maxOutputChannels') > 0:
            devices.append(item)
    return devices


if __name__ == '__main__':
    # get_audio_devices()
    print('输入设备:')
    for item in get_audio_input_devices():
        print(item)
    print('输出设备:')
    for item in get_audio_output_devices():
        print(item)
